import re
from typing import TypeGuard

from bs4 import BeautifulSoup
from bs4.element import NavigableString, Tag

from labels.model.file import Location, LocationReadCloser
from labels.model.package import Package
from labels.model.relationship import Relationship
from labels.model.release import Environment
from labels.model.resolver import Resolver
from labels.parsers.cataloger.dotnet.package_builder import new_dotnet_package
from labels.parsers.cataloger.utils import get_enriched_location

RUNTIME_TAG_NAME = "supportedruntime"
RUNTIME_TAG_PARENT_NAME = "startup"
RUNTIME_TAG_SKU_ATTRIBUTE = "sku"
RUNTIME_TAG_DEPENDENCY_PATTERN = re.compile(r".NETFramework,Version=v(?P<version>[^\s,]*)")

EXECUTABLE_PACKAGE_NAME = "netframework"


def parse_dotnet_config_executable(
    _resolver: Resolver | None,
    _environment: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    file_content = BeautifulSoup(reader.read_closer.read(), features="html.parser")

    packages = _collect_packages(file_content, reader.location)

    return packages, []


def _collect_packages(file_content: BeautifulSoup, location: Location) -> list[Package]:
    packages: list[Package] = []

    net_runtime_tag = file_content.find(RUNTIME_TAG_NAME)

    if _is_valid_runtime_tag(net_runtime_tag):
        runtime_info = net_runtime_tag.get(RUNTIME_TAG_SKU_ATTRIBUTE, "")
        version_match = RUNTIME_TAG_DEPENDENCY_PATTERN.match(str(runtime_info))
        if not runtime_info or not version_match:
            return packages

        version = version_match.group("version")

        new_location = get_enriched_location(
            location, line=net_runtime_tag.sourceline, is_transitive=False, is_dev=False
        )

        new_package = new_dotnet_package(EXECUTABLE_PACKAGE_NAME, version, new_location)
        if new_package:
            packages.append(new_package)

    return packages


def _is_valid_runtime_tag(
    net_runtime_tag: Tag | NavigableString | None,
) -> TypeGuard[Tag]:
    return bool(
        isinstance(net_runtime_tag, Tag)
        and isinstance(net_runtime_tag.parent, Tag)
        and net_runtime_tag.parent.name == RUNTIME_TAG_PARENT_NAME
    )



import re
from collections.abc import Iterable

from gemfileparser import Dependency, GemfileParser

from labels.model.file import Location, LocationReadCloser
from labels.model.package import Package
from labels.model.relationship import Relationship
from labels.model.release import Environment
from labels.model.resolver import Resolver
from labels.parsers.cataloger.ruby.package_builder import new_gem_package
from labels.parsers.cataloger.utils import get_enriched_location

GEMFILE_DEP: re.Pattern[str] = re.compile(
    r'^\s*(?P<gem>gem [\'"].*?[\'"],?( [\'"][><~=]{0,2}\s?[\d\.]+[\'"],?){0,2})',
)
NOT_PROD_DEP: re.Pattern[str] = re.compile(
    r":group => \[?[:\w\-, ]*(:development|:test)",
)
NOT_PROD_GROUP: re.Pattern[str] = re.compile(r"(\s*)group :(test|development)")
GEM_LOCK_DEP: re.Pattern[str] = re.compile(
    r"^\s{4}(?P<gem>[^\s]*)\s\([^\d]*(?P<version>.*)\)$",
)


def parse_gemfile(
    _resolver: Resolver | None,
    _environment: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    packages = _collect_packages(reader.read_closer.read(), reader)

    return packages, []


def _collect_packages(file_content: str, reader: LocationReadCloser) -> list[Package]:
    packages: list[Package] = []
    in_group_block = False
    end_line = ""

    for line_number, line in enumerate(file_content.splitlines(), 1):
        if in_group_block:
            if line == end_line:
                in_group_block = False
            elif (
                (matched := GEMFILE_DEP.search(line))
                and matched
                and (dependency := _create_dependency(line_number, reader.location, matched, line))
            ):
                packages.append(dependency)
            continue

        if match_group := NOT_PROD_GROUP.search(line):
            in_group_block = True
            end_line = f"{match_group.group(1)}end"
            continue

        if (
            (matched := GEMFILE_DEP.search(line))
            and matched
            and (dependency := _create_dependency(line_number, reader.location, matched, line))
        ):
            packages.append(dependency)

    return packages


def _create_dependency(
    line_number: int,
    location: Location,
    matched: re.Match[str],
    line: str,
) -> Package | None:
    gem_info = GemfileParser.preprocess(matched.group("gem"))[3:]
    product, version = _parse_line(gem_info)

    is_dev = _is_dev_dependency(line)

    new_location = get_enriched_location(
        location, line=line_number, is_dev=is_dev, is_transitive=False
    )

    return new_gem_package(product, version, new_location)


def _parse_line(in_line: str) -> tuple[str, str]:
    line: list[str] = []
    line = in_line.split(",")

    column_list: list[str] = []
    for column in line:
        stripped_column = (
            column.replace("'", "")
            .replace('"', "")
            .replace("%q<", "")
            .replace("(", "")
            .replace(")", "")
            .replace("[", "")
            .replace("]", "")
            .strip()
        )
        column_list.append(stripped_column)

    deps = _match_dep_criteria(column_list)
    deps_dict = deps.to_dict()
    product: str = deps_dict["name"]
    version: str = _format_requirements(deps_dict["requirement"])
    return product, version


def _match_dep_criteria(
    column_list: Iterable[str],
) -> Dependency:
    dep = Dependency()
    for column in column_list:
        for criteria, criteria_regex in GemfileParser.gemfile_regexes.items():
            match = criteria_regex.match(column)
            if match:
                if criteria == "requirement":
                    dep.requirement.append(match.group(criteria))
                else:
                    setattr(dep, criteria, match.group(criteria))
                break
    return dep


def _format_requirements(requirements: list[str]) -> str:
    formatted: str = ""
    if len(requirements) == 0:
        return formatted
    requirements = [req.replace(" ", "") for req in requirements]
    first_req = requirements[0]
    if "~" in first_req:
        if len(requirements) == 1:
            dot_times = first_req.count(".")
            if dot_times <= 1:
                formatted = first_req.replace("~>", "^")
            else:
                formatted = first_req.replace("~>", "~")
        elif len(requirements) == 2:
            sec_req = requirements[1]
            if ">=" in sec_req:
                formatted = sec_req.replace(">=", "^")
            elif "!=" in sec_req:
                sec_ver = sec_req.replace("!=", "")
                first_ver = first_req.replace("!=", "")
                formatted = f">={first_ver} <{sec_ver}  || >{sec_ver}"
    else:
        formatted = " ".join(requirements)
    return formatted


def _is_dev_dependency(line: str) -> bool:
    return bool(NOT_PROD_DEP.search(line))



from datetime import datetime

import pydantic

from labels.enrichers.golang.get import fetch_latest_version_info
from labels.model.metadata import HealthMetadata
from labels.model.package import Package


class GolangModuleEntry(pydantic.BaseModel):
    h1_digest: str


def complete_package(package: Package) -> Package:
    latest = fetch_latest_version_info(package.name)
    if not latest:
        return package
    package.health_metadata = HealthMetadata(
        latest_version=latest["Version"],
        latest_version_created_at=datetime.fromisoformat(latest["Time"]),
        artifact=None,
    )
    return package



import re

from labels.model.ecosystem_data.alpine import ApkDBEntry
from labels.model.file import Location, LocationReadCloser
from labels.model.package import Package
from labels.model.relationship import Relationship, RelationshipType
from labels.model.release import Environment
from labels.model.resolver import Resolver
from labels.parsers.cataloger.alpine.package_builder import new_alpine_package

APK_DB_GLOB = "**/lib/apk/db/installed"


def parse_apk_db(
    _resolver: Resolver | None,
    environment: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    apk_db_entries = _collect_apk_db_entries(reader.read_closer.read())

    packages = _collect_packages(apk_db_entries, environment, reader.location)
    relationships = _collect_relationships(packages)

    return packages, relationships


def _collect_apk_db_entries(content: str) -> list[ApkDBEntry]:
    apk_db_entries: list[ApkDBEntry] = []
    stripped_content = content.strip().split("\n\n")

    for package_content in stripped_content:
        if not package_content:
            continue

        apk_db_entry = _parse_package_content(package_content)
        apk_db_entries.append(apk_db_entry)

    return apk_db_entries


def _parse_package_content(content: str) -> ApkDBEntry:
    parsed_data_dict: dict[str, str] = {}
    lines = content.split("\n")
    key = ""

    for line in lines:
        key = _process_line_and_update_data(line, key, parsed_data_dict)

    return _construct_apk_db_entry(parsed_data_dict)


def _process_line_and_update_data(line: str, key: str, data: dict[str, str]) -> str:
    if ":" in line:
        key, value = line.split(":", 1)
        data[key] = value
    elif key and key in data:
        data[key] += "\n" + line.strip()
    return key


def _construct_apk_db_entry(data: dict[str, str]) -> ApkDBEntry:
    return ApkDBEntry(
        package=data.get("P"),
        version=data.get("V"),
        origin_package=data.get("o"),
        maintainer=data.get("m"),
        architecture=data.get("A"),
        dependencies=_parse_list_values(data.get("D")),
        provides=_parse_list_values(data.get("p")),
    )


def _parse_list_values(value: str | None, delimiter: str | None = None) -> list[str]:
    delimiter = delimiter or " "
    if not value:
        return []

    return value.split(delimiter)


def _collect_packages(
    apk_db_entries: list[ApkDBEntry], environment: Environment | None, location: Location
) -> list[Package]:
    packages: list[Package] = []
    release = environment.linux_release if environment else None

    for apk_db_entry in apk_db_entries:
        package = new_alpine_package(apk_db_entry, release, location)
        if package:
            packages.append(package)

    return packages


def _collect_relationships(packages: list[Package]) -> list[Relationship]:
    relationships: list[Relationship] = []
    lookup = _build_lookup_table(packages)

    for package in packages:
        if not isinstance(package.ecosystem_data, ApkDBEntry):  # pragma: no cover
            continue

        package_apk_db_entry = package.ecosystem_data
        for dep_specifier in package_apk_db_entry.dependencies:
            dependency_name = _strip_version_specifier(dep_specifier)
            relationships.extend(
                Relationship(
                    from_=dependency_package.id_,
                    to_=package.id_,
                    type=RelationshipType.DEPENDENCY_OF_RELATIONSHIP,
                )
                for dependency_package in lookup.get(dependency_name, [])
            )

    return relationships


def _build_lookup_table(packages: list[Package]) -> dict[str, list[Package]]:
    lookup: dict[str, list[Package]] = {}

    for package in packages:
        if not isinstance(package.ecosystem_data, ApkDBEntry):  # pragma: no cover
            continue

        apkg = package.ecosystem_data
        if package.name not in lookup:
            lookup[package.name] = [package]
        else:
            lookup[package.name].append(package)

        for provides in apkg.provides:
            provides_k = _strip_version_specifier(provides)
            if provides_k not in lookup:
                lookup[provides_k] = [package]
            else:
                lookup[provides_k].append(package)

    return lookup


def _strip_version_specifier(version: str) -> str:
    splitted_version: list[str] = re.split("[<>=]", version)
    return splitted_version[0]



from typing import NamedTuple

from labels.model.file import Location, LocationReadCloser
from labels.model.indexables import IndexedDict, IndexedList, ParsedValue
from labels.model.package import Package
from labels.model.relationship import Relationship, RelationshipType
from labels.model.release import Environment
from labels.model.resolver import Resolver
from labels.parsers.cataloger.swift.package_builder import new_cocoa_pods_package
from labels.parsers.cataloger.utils import get_enriched_location
from labels.parsers.collection.yaml import parse_yaml_with_tree_sitter


class PodInfo(NamedTuple):
    name: str
    version: str
    dependencies: list[str]
    is_transitive: bool
    line: int


def parse_podfile_lock(
    _resolver: Resolver | None,
    _environment: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    file_content = _safe_parse_podfile(reader.read_closer.read())
    if not isinstance(file_content, IndexedDict):
        return [], []

    pods_info = _parse_pods_info(file_content)
    if not pods_info:
        return [], []

    packages = _collect_packages(pods_info, reader.location)
    relationships = _generate_relations(pods_info, packages)

    return packages, relationships


def _safe_parse_podfile(content: str) -> ParsedValue | None:
    try:
        parsed_podfile = parse_yaml_with_tree_sitter(content)
    except ValueError:
        return None

    return parsed_podfile


def _parse_pods_info(file_content: IndexedDict[str, ParsedValue]) -> list[PodInfo]:
    pods_info: list[PodInfo] = []

    direct_dependencies = file_content.get("DEPENDENCIES")
    pods = file_content.get("PODS")
    if not isinstance(direct_dependencies, IndexedList) or not isinstance(pods, IndexedList):
        return pods_info

    for index, pod in enumerate(pods):
        pod_name, pod_version = _extract_pod_info(pod)
        pod_dependencies = _extract_pod_dependencies(pod)
        is_transitive = pod_name not in direct_dependencies
        line = pods.get_position(index).start.line

        pods_info.append(
            PodInfo(
                name=pod_name,
                version=pod_version,
                dependencies=pod_dependencies,
                is_transitive=is_transitive,
                line=line,
            )
        )

    return pods_info


def _collect_packages(pods_info: list[PodInfo], location: Location) -> list[Package]:
    packages: list[Package] = []

    for pod_info in pods_info:
        new_location = get_enriched_location(
            location, is_transitive=pod_info.is_transitive, line=pod_info.line
        )

        package = new_cocoa_pods_package(
            name=pod_info.name, version=pod_info.version, location=new_location
        )
        if package:
            packages.append(package)

    return packages


def _generate_relations(pods_info: list[PodInfo], packages: list[Package]) -> list[Relationship]:
    relationships: list[Relationship] = []

    packages_by_name = {package.name: package for package in packages}

    for pod_info in pods_info:
        package = packages_by_name.get(pod_info.name)
        if not package:
            continue

        relationships.extend(
            Relationship(
                from_=package_dependency.id_,
                to_=package.id_,
                type=RelationshipType.DEPENDENCY_OF_RELATIONSHIP,
            )
            for dependency in pod_info.dependencies
            if (package_dependency := packages_by_name.get(dependency))
        )

    return relationships


def _extract_pod_info(pod: ParsedValue) -> tuple[str, str]:
    if isinstance(pod, str | IndexedDict):
        pod_blob = pod if isinstance(pod, str) else next(iter(pod))
        pod_name = pod_blob.split(" ")[0]
        pod_version = pod_blob.split(" ")[1].strip("()")
        return pod_name, pod_version
    return "", ""


def _extract_pod_dependencies(pod: ParsedValue) -> list[str]:
    dependencies = []
    if isinstance(pod, IndexedDict):
        for value in pod.values():
            if isinstance(value, IndexedList):
                for dep in value:
                    if isinstance(dep, str):
                        dep_name = dep.split(" ")[0]
                        dependencies.append(dep_name)
    return dependencies



from packageurl import PackageURL
from pydantic import ValidationError

from labels.model.ecosystem_data.alpine import ApkDBEntry
from labels.model.file import Location
from labels.model.package import Language, Package, PackageType
from labels.model.release import Release
from labels.parsers.cataloger.utils import get_enriched_location, log_malformed_package_warning


def new_alpine_package(
    data: ApkDBEntry,
    release: Release | None,
    db_location: Location,
) -> Package | None:
    name = data.package
    version = data.version

    if not name or not version:
        return None

    new_location = get_enriched_location(db_location)

    p_url = _get_alpine_package_url(data, release)

    try:
        return Package(
            name=name,
            version=version,
            locations=[new_location],
            p_url=p_url,
            type=PackageType.ApkPkg,
            ecosystem_data=data,
            found_by=None,
            health_metadata=None,
            language=Language.UNKNOWN_LANGUAGE,
        )
    except ValidationError as ex:
        log_malformed_package_warning(new_location, ex)
        return None


def _get_alpine_package_url(entry: ApkDBEntry, distro: Release | None) -> str:
    qualifiers = {"arch": entry.architecture or ""} if entry else {}

    if entry and entry.origin_package != entry.package and entry.origin_package:
        qualifiers["upstream"] = entry.origin_package
    distro_qualifiers = []

    if distro and distro.id_:
        qualifiers["distro_id"] = distro.id_
        distro_qualifiers.append(distro.id_)

    if distro and distro.version_id:
        qualifiers["distro_version_id"] = distro.version_id
        distro_qualifiers.append(distro.version_id)
    elif distro and distro.build_id:
        distro_qualifiers.append(distro.build_id)

    if distro_qualifiers:
        qualifiers["distro"] = "-".join(distro_qualifiers)

    return PackageURL(  # type: ignore[misc]
        type="apk",
        namespace=distro.id_.lower() if distro and distro.id_ else "",
        name=entry.package,
        version=entry.version,
        qualifiers=qualifiers,
    ).to_string()



import logging
import struct
from pathlib import Path
from typing import BinaryIO

from labels.model.file import Location, LocationReadCloser
from labels.model.package import Package
from labels.model.relationship import Relationship
from labels.model.release import Environment
from labels.model.resolver import Resolver
from labels.parsers.cataloger.redhat.package_builder import new_redhat_package
from labels.parsers.cataloger.redhat.rpmdb.domain.entry import IndexEntry, header_import
from labels.parsers.cataloger.redhat.rpmdb.domain.package import get_nevra

LOGGER = logging.getLogger(__name__)


def parse_rpm_file(
    _: Resolver,
    env: Environment,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    """Parse an individual .rpm file to extract package metadata.

    An RPM file structure:
    - Lead (96 bytes) - legacy header
    - Signature Header - signatures and checksums
    - Header - package metadata (what we need)
    - Payload - compressed files
    """
    packages = _collect_packages(reader.location, env)

    return packages, []


def _collect_packages(location: Location, env: Environment) -> list[Package]:
    packages: list[Package] = []

    if not location.coordinates:
        return packages

    try:
        with Path(location.coordinates.real_path).open("rb") as rpm_file:
            rpm_file.seek(96)  # Skip the lead (96 bytes)

            index_entries = _parse_rpm_header(rpm_file)
            if not index_entries:
                return packages

            package_info = get_nevra(index_entries)
            package = new_redhat_package(entry=package_info, env=env, location=location)
            if package:
                packages.append(package)

    except (OSError, ValueError, struct.error) as ex:
        LOGGER.warning("Failed to parse RPM file %s: %s", location.coordinates.real_path, ex)

    return packages


def _parse_rpm_header(f: BinaryIO) -> list[IndexEntry] | None:
    if not _skip_signature_section(f):
        return None

    header_data = _read_main_header_block(f)

    if not header_data:
        return None

    return header_import(header_data)


def _skip_signature_section(f: BinaryIO) -> bool:
    magic = f.read(8)
    if len(magic) < 8 or not _has_rpm_header_magic(magic):
        return False

    index_length = _read_u32_be(f)
    data_length = _read_u32_be(f)
    f.seek(index_length * 16 + data_length, 1)
    _align_to_8_bytes(f)

    return True


def _align_to_8_bytes(f: BinaryIO) -> None:
    pos = f.tell()
    f.seek((8 - (pos % 8)) % 8, 1)


def _read_main_header_block(f: BinaryIO) -> bytes | None:
    magic = f.read(8)
    if len(magic) < 8 or not _has_rpm_header_magic(magic):
        return None

    index_length = _read_u32_be(f)
    data_length = _read_u32_be(f)
    total = 8 + (index_length * 16) + data_length
    f.seek(-8, 1)

    data = f.read(total)

    return data if len(data) >= total else None


def _read_u32_be(f: BinaryIO) -> int:
    return int.from_bytes(f.read(4), "big")


def _has_rpm_header_magic(buf: bytes) -> bool:
    return len(buf) >= 3 and buf[:3] == b"\x8e\xad\xe8"



import re

import packaging.version
import pefile

from labels.model.file import Location, LocationReadCloser
from labels.model.package import Package
from labels.model.relationship import Relationship
from labels.model.release import Environment
from labels.model.resolver import Resolver
from labels.parsers.cataloger.dotnet.package_builder import new_dotnet_package

SPACE_REGEX = re.compile(r"[\s]+")
NUMBER_REGEX = re.compile(r"\d")
VERSION_PUNCTUATION_REGEX = re.compile(r"[.,]+")


def parse_dotnet_portable_executable(
    _resolver: Resolver | None,
    _environment: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    pe_representation = _safe_parse_version_pe_file(reader)
    if not pe_representation:
        return [], []

    packages = _collect_packages(pe_representation, reader.location)

    return packages, []


def _safe_parse_version_pe_file(reader: LocationReadCloser) -> pefile.PE | None:
    if not reader.location.coordinates:
        return None

    try:
        pe_representation = pefile.PE(reader.location.coordinates.real_path, fast_load=False)
    except pefile.PEFormatError:
        return None

    return pe_representation


def _collect_packages(pe_representation: pefile.PE, location: Location) -> list[Package]:
    packages: list[Package] = []

    version_resource = _parse_version_resource(pe_representation)
    if not version_resource:
        return packages

    name = _find_name(version_resource)
    version = _find_version(version_resource)

    dotnet_package = new_dotnet_package(name, version, location)
    if dotnet_package:
        packages.append(dotnet_package)

    return packages


def _parse_version_resource(portable: pefile.PE) -> dict[str, str] | None:
    if not hasattr(portable, "VS_VERSIONINFO"):
        return None

    for idx, _ in enumerate(portable.VS_VERSIONINFO):
        if not hasattr(portable, "FileInfo") or len(portable.FileInfo) <= idx:
            continue

        parsed_file_info = _process_file_info(portable.FileInfo[idx])
        if parsed_file_info:
            return parsed_file_info

    return None


def _process_file_info(
    file_info: list["pefile._StringFileInfo"],
) -> dict[str, str] | None:
    for entry in file_info:
        if not hasattr(entry, "StringTable"):
            continue

        parsed_file_info = _process_string_table(entry.StringTable)
        if parsed_file_info:
            return parsed_file_info

    return None


def _process_string_table(
    string_table: list["pefile._StringTable"],
) -> dict[str, str]:
    string_table_dict: dict[str, str] = {}

    for string_table_entry in string_table:
        string_table_dict["LangID"] = string_table_entry.LangID.decode("utf-8")
        for key, value in string_table_entry.entries.items():
            string_table_dict[key.decode("utf-8")] = value.decode("utf-8")

    return string_table_dict


def _find_name(version_resources: dict[str, str]) -> str:
    name_fields = [
        "ProductName",
        "FileDescription",
        "InternalName",
        "OriginalFilename",
    ]

    if _is_microsoft(version_resources):
        name_fields = [
            "FileDescription",
            "InternalName",
            "OriginalFilename",
            "ProductName",
        ]

    for field in name_fields:
        value = _space_normalize(version_resources.get(field, ""))
        if value:
            return value

    return ""


def _is_microsoft(version_resources: dict[str, str]) -> bool:
    company_name = version_resources.get("CompanyName", "").lower()
    product_name = version_resources.get("ProductName", "").lower()

    return "microsoft" in company_name or "microsoft" in product_name


def _space_normalize(value: str) -> str:
    value = value.strip()

    if value == "":
        return ""

    value = value.encode("utf-8", "replace").decode("utf-8")

    value = SPACE_REGEX.sub(" ", value)

    value = re.sub(r"[\x00-\x1f]", "", value)

    value = SPACE_REGEX.sub(" ", value)

    return value.strip()


def _find_version(version_resources: dict[str, str]) -> str:
    raw_product_version = version_resources.get("ProductVersion", "")
    raw_file_version = version_resources.get("FileVersion", "")

    product_version = _extract_version(raw_product_version)
    file_version = _extract_version(raw_file_version)

    if not product_version and not file_version:
        return ""

    if not product_version:
        return file_version

    if not file_version:
        return product_version

    semantic_choice = _keep_greater_semantic_version(product_version, file_version)
    if semantic_choice:
        return semantic_choice

    return _choose_by_detail_and_numbers(product_version, file_version)


def _choose_by_detail_and_numbers(product_version: str, file_version: str) -> str:
    product_detail = _punctuation_count(product_version)
    file_detail = _punctuation_count(file_version)

    has_num_product = _contains_number(product_version)
    has_num_file = _contains_number(file_version)

    if has_num_file and file_detail > 0 and (not has_num_product or file_detail > product_detail):
        return file_version

    if has_num_product:
        return product_version
    if has_num_file:
        return file_version

    return product_version or file_version


def _keep_greater_semantic_version(product_version: str, file_version: str) -> str:
    try:
        semantic_product_version = packaging.version.parse(product_version)
    except ValueError:
        return ""

    try:
        semantic_file_version = packaging.version.parse(file_version)
    except ValueError:
        return product_version

    if semantic_product_version == semantic_file_version:
        return ""

    if semantic_file_version > semantic_product_version:
        return file_version

    return product_version


def _punctuation_count(string: str) -> int:
    return len(VERSION_PUNCTUATION_REGEX.findall(string))


def _extract_version(version: str) -> str:
    version = version.strip()

    result = ""

    for index, character in enumerate(version.split()):
        if _contains_number(result) and not _contains_number(character):
            return result

        if index == 0:
            result = character
        else:
            result += " " + character

    return result


def _contains_number(string: str) -> bool:
    return any(character.isdigit() for character in string)



from typing import cast

from labels.model.file import Location, LocationReadCloser
from labels.model.indexables import IndexedDict, IndexedList, ParsedValue
from labels.model.package import Package
from labels.model.relationship import Relationship, RelationshipType
from labels.model.release import Environment
from labels.model.resolver import Resolver
from labels.parsers.cataloger.php.package_builder import new_package_from_composer
from labels.parsers.cataloger.utils import get_enriched_location
from labels.parsers.collection.json import parse_json_with_tree_sitter

EMPTY_LIST: IndexedList[ParsedValue] = IndexedList()


def parse_installed_json(
    _resolver: Resolver | None,
    _environment: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    file_content = parse_json_with_tree_sitter(reader.read_closer.read())
    if not isinstance(file_content, (IndexedDict, IndexedList)):  # pragma: no cover
        return [], []

    packages = _extract_packages(file_content, reader.location)
    relationships = _extract_relationships(packages, file_content)

    return packages, relationships


def _extract_packages(
    file_content: IndexedDict[str, ParsedValue] | IndexedList[ParsedValue],
    location: Location,
) -> list[Package]:
    packages = []
    dev_packages = EMPTY_LIST

    if (
        isinstance(file_content, IndexedDict)
        and "dev-package-names" in file_content
        and isinstance(file_content["dev-package-names"], IndexedList)
    ):
        dev_packages = file_content["dev-package-names"]

    packages_list = (
        file_content["packages"]
        if isinstance(file_content, IndexedDict) and "packages" in file_content
        else file_content
    )

    for package in packages_list if isinstance(packages_list, IndexedList) else EMPTY_LIST:
        if not isinstance(package, IndexedDict):
            continue

        name = cast("str", package.get("name"))
        version = cast("str", package.get("version"))

        if not name:
            continue

        is_dev = _is_dev_package(package, dev_packages)
        new_location = get_enriched_location(
            location,
            line=package.get_key_position("name").start.line,
            is_dev=is_dev,
            is_transitive=False,
        )

        pkg_item = new_package_from_composer(name=name, version=version, location=new_location)
        if pkg_item:
            packages.append(pkg_item)

    return packages


def _extract_relationships(
    packages: list[Package], file_content: IndexedDict[str, ParsedValue] | IndexedList[ParsedValue]
) -> list[Relationship]:
    relationships = []
    packages_list = (
        file_content["packages"]
        if isinstance(file_content, IndexedDict) and "packages" in file_content
        else file_content
    )

    if not isinstance(packages_list, IndexedList):
        return []

    for package in packages_list:
        parsed_package_with_dependencies = _get_parsed_package_and_dependencies(package, packages)
        if not parsed_package_with_dependencies:
            continue

        parsed_package, dependencies = parsed_package_with_dependencies
        if not parsed_package or not dependencies:
            continue

        raw_dependency_names = list(dependencies.keys()) if dependencies else []
        for dep_name in raw_dependency_names:
            dependency_parsed_package = next((x for x in packages if x.name == dep_name), None)
            if dependency_parsed_package:
                relationships.append(
                    Relationship(
                        from_=dependency_parsed_package.id_,
                        to_=parsed_package.id_,
                        type=RelationshipType.DEPENDENCY_OF_RELATIONSHIP,
                    ),
                )

    return relationships


def _is_dev_package(
    package: IndexedDict[str, ParsedValue], dev_packages: IndexedList[ParsedValue]
) -> bool:
    name_value = package.get("name")
    name = name_value if isinstance(name_value, str) else None

    return name in dev_packages if name else False


def _get_parsed_package_and_dependencies(
    package_data: ParsedValue,
    packages: list[Package],
) -> tuple[Package | None, IndexedDict[str, ParsedValue]] | None:
    if not isinstance(package_data, IndexedDict):
        return None

    package_name = package_data.get("name")
    if not isinstance(package_name, str):
        return None

    parsed_package = next((p for p in packages if p.name == package_name), None)

    raw_dependencies = package_data.get("require")
    if not parsed_package or not isinstance(raw_dependencies, IndexedDict):
        return None

    return parsed_package, raw_dependencies



from collections.abc import Callable

from labels.model.package import Package, PackageType
from labels.model.syft_sbom import SyftArtifact
from labels.parsers.builder.alpine.builder_apk import builder as builder_apk_pkg
from labels.parsers.builder.arch.builder_alpm import builder as builder_alpm_pkg
from labels.parsers.builder.debian.builder_deb import builder as builder_deb_pkg
from labels.parsers.builder.generic_builder import builder as generic_builder
from labels.parsers.builder.java.builder_java import builder as builder_java_pkg
from labels.parsers.builder.python.builder_python import builder as builder_python_pkg
from labels.parsers.builder.redhat.builder_rpm import builder as builder_rpmdb_pkg

BUILDERS: dict[PackageType, Callable[[SyftArtifact, PackageType], Package | None]] = {
    PackageType.DebPkg: builder_deb_pkg,
    PackageType.ApkPkg: builder_apk_pkg,
    PackageType.AlpmPkg: builder_alpm_pkg,
    PackageType.RpmPkg: builder_rpmdb_pkg,
    PackageType.PythonPkg: builder_python_pkg,
    PackageType.JavaPkg: builder_java_pkg,
    PackageType.NpmPkg: generic_builder,
    PackageType.DotnetPkg: generic_builder,
    PackageType.GemPkg: generic_builder,
    PackageType.PhpComposerPkg: generic_builder,
    PackageType.PhpPearPkg: generic_builder,
    PackageType.DartPubPkg: generic_builder,
    PackageType.SwiftPkg: generic_builder,
}


def build_package(
    artifact: SyftArtifact,
    package_type: PackageType,
) -> Package | None:
    if builder := BUILDERS.get(package_type):
        return builder(artifact, package_type)
    return None



import functools
import hashlib
import io
import os
import pickle  # nosec
import socket
from collections.abc import Callable
from contextlib import suppress
from typing import Any, TypeVar, cast

import diskcache
import diskcache.core
from diskcache import Lock
from platformdirs import user_cache_dir
from redis import Redis

REDIS_CACHE_ENDPOINT = os.environ.get("CACHE_URL")
REDIS_CLIENT = None

TVar = TypeVar("TVar")
TFun = TypeVar("TFun", bound=Callable[..., Any])

if REDIS_CACHE_ENDPOINT:
    try:
        host, _port = REDIS_CACHE_ENDPOINT.split(":", maxsplit=1)
        port = int(_port)

        # Try to connect to redis endpoint
        socket.create_connection((host, port), timeout=2)

        REDIS_CLIENT = Redis(
            host=host,
            port=port,
            ssl=True,
            username="batch-rw-user",
            password=os.environ.get("CACHE_BATCH_RW_USER_PASSWORD"),
        )
    except (OSError, ValueError):
        REDIS_CLIENT = None

DISK_CACHE = diskcache.Cache(user_cache_dir("fluid-labels", "fluidattacks"))


# 4 weeks
TTL = 604800 * 4

# Whitelist of allowed modules and classes for unpickling
_PICKLE_ALLOWED_CLASSES: frozenset[tuple[str, str]] = frozenset(
    {
        # Built-in types
        ("builtins", "dict"),
        ("builtins", "list"),
        ("builtins", "set"),
        ("builtins", "frozenset"),
        ("builtins", "tuple"),
        ("builtins", "bytes"),
        ("builtins", "bytearray"),
        # Collections
        ("collections", "OrderedDict"),
        ("collections", "defaultdict"),
        # requests.Response and its dependencies
        ("requests.models", "Response"),
        ("requests.structures", "CaseInsensitiveDict"),
        ("requests.cookies", "RequestsCookieJar"),
        ("http.cookiejar", "CookieJar"),
        ("http.cookiejar", "Cookie"),
        ("urllib3.response", "HTTPResponse"),
        ("urllib3._collections", "HTTPHeaderDict"),
        # datetime types
        ("datetime", "datetime"),
        ("datetime", "date"),
        ("datetime", "time"),
        ("datetime", "timedelta"),
    }
)


class RestrictedUnpickler(pickle.Unpickler):
    """Unpickler that only allows specific safe classes."""

    def find_class(self, module: str, name: str) -> type:
        if (module, name) in _PICKLE_ALLOWED_CLASSES:
            return super().find_class(module, name)

        err_msg = f"Forbidden class in cache: {module}.{name}"
        raise pickle.UnpicklingError(err_msg)


def safe_pickle_loads(data: bytes) -> Any:  # noqa: ANN401
    return RestrictedUnpickler(io.BytesIO(data)).load()


def make_hashable(item: Any) -> str:  # noqa: ANN401
    serialized_object = pickle.dumps(item)

    return hashlib.sha256(serialized_object).hexdigest()


def generate_cache_key(func: Any, args: Any, kwargs: dict[Any, Any]) -> str:  # noqa: ANN401
    key = f"{func.__module__}.{func.__name__}-"
    key += str(make_hashable((args, kwargs)))
    return hashlib.sha256(key.encode()).hexdigest()


def dual_cache(func: TVar) -> TVar:
    _func = cast("Callable[..., Any]", func)

    @functools.wraps(_func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:  # noqa: ANN401
        if os.environ.get("SKIP_FA_LABELS_CACHE_USE"):
            return _func(*args, **kwargs)

        cache_key = generate_cache_key(_func, args, kwargs)

        # Try to recover from disk cache
        if value := DISK_CACHE.get(cache_key):
            return value
        # Try to recover from Redis cache (ElastiCache) if available
        if REDIS_CLIENT:
            cached_result = REDIS_CLIENT.get(cache_key)

            if cached_result:
                result = safe_pickle_loads(cached_result)
                with Lock(DISK_CACHE, "dual_cache"):
                    DISK_CACHE.set(
                        cache_key,
                        result,
                        expire=TTL,
                        retry=True,
                    )  # Save to disk with TTL

                return result

        # Run the function and store the result in both caches
        result = _func(*args, **kwargs)
        with suppress(diskcache.core.Timeout), Lock(DISK_CACHE, "dual_cache"):
            DISK_CACHE.set(cache_key, result, expire=TTL, retry=True)  # Save to disk with TTL
        if REDIS_CLIENT:
            REDIS_CLIENT.setex(cache_key, TTL, pickle.dumps(result))  # Save to Redis with TTL
        return result

    return cast("TVar", wrapper)



import json
import logging
from datetime import datetime
from enum import Enum
from typing import Any

from pydantic import BaseModel

from labels.model.advisories import Advisory
from labels.model.indexables import IndexedDict, IndexedList
from labels.model.package import Package
from labels.model.relationship import Relationship

LOGGER = logging.getLogger(__name__)


class EnumEncoder(json.JSONEncoder):
    def default(self, item: Any) -> Any:  # noqa: ANN401
        if isinstance(item, Enum):
            return item.value
        if isinstance(item, IndexedList):
            return list(item.data)
        if isinstance(item, IndexedDict):
            return dict(item.data)
        if isinstance(item, BaseModel):
            return item.model_dump()
        if isinstance(item, datetime):
            return item.isoformat()
        return super().default(item)


def serialize_package(package: Package) -> dict[str, Any]:
    return {
        "id": package.id_,
        "name": package.name,
        "version": package.version,
        "locations": [
            {
                "path": loc.path(),
                "line": loc.coordinates.line if loc.coordinates and loc.coordinates.line else None,
                "layer": loc.coordinates.file_system_id
                if loc.coordinates and loc.coordinates.file_system_id
                else None,
                "dependency_type": loc.output_dependency_type(),
                "scope": loc.scope.value,
                "reachable_cves": loc.reachable_cves,
                "location_id": loc.location_id(),
                "top_parents": loc.top_parents if loc.top_parents is not None else None,
                "dependency_chains": loc.dependency_chains
                if loc.dependency_chains is not None
                else None,
            }
            for loc in package.locations
        ],
        "type": package.type.value,
        "language": package.language.value,
        "platform": package.type.get_platform_value(),
        "package_url": package.p_url,
        "found_by": package.found_by,
        "health_metadata": package.health_metadata.model_dump()
        if package.health_metadata
        else None,
        "safe_versions": package.safe_versions,
        "advisories": [serialize_advisory(adv) for adv in package.advisories or []],
    }


def serialize_advisory(advisory: Advisory) -> dict[str, Any]:
    return {
        "cpes": advisory.cpes,
        "description": advisory.details,
        "epss": advisory.epss,
        "id": advisory.id,
        "namespace": advisory.package_manager,
        "percentile": advisory.percentile,
        "severity": advisory.severity_level,
        "urls": [advisory.source],
        "version_constraint": advisory.vulnerable_version,
        "platform_version": advisory.platform_version,
        "fixed_versions": advisory.fixed_versions,
        "fix_metadata": advisory.fix_metadata.model_dump(exclude_none=True)
        if advisory.fix_metadata
        else None,
        "fix_metadata_new": advisory.fix_metadata_new.model_dump(exclude_none=True)
        if advisory.fix_metadata_new
        else None,
        "cvss4": advisory.severity_v4,
        "cwe_ids": advisory.cwe_ids,
        "cve_finding": advisory.cve_finding,
        "auto_approve": advisory.auto_approve,
        "upstream_package": advisory.upstream_package,
        "kev_catalog": advisory.kev_catalog,
    }


def serialize_packages(packages: list[Package]) -> list[dict[str, Any]]:
    raw_pkgs = [serialize_package(pkg) for pkg in packages]
    return validate_pkgs(raw_pkgs)


def validate_pkgs(raw_sbom_pkgs: list[dict[str, Any]]) -> list[dict[str, Any]]:
    pkgs_to_remove = []
    for index, raw_pkg in enumerate(raw_sbom_pkgs):
        try:
            pkg = json.loads(json.dumps(raw_pkg, cls=EnumEncoder))
            raw_sbom_pkgs[index] = pkg
        except TypeError as err:
            LOGGER.warning(
                "Cannot serialize package at index %d: %s. Error: %s",
                index,
                raw_pkg,
                str(err),
            )
            pkgs_to_remove.append(index)

    for index in sorted(pkgs_to_remove, reverse=True):
        raw_sbom_pkgs.pop(index)
    return raw_sbom_pkgs


def build_relationship_map(
    relationships: list[Relationship],
) -> list[dict[str, str | list[str]]]:
    grouped: dict[tuple[str, str], set[str]] = {}

    for rel in relationships:
        if rel.type.value == "dependency-of":
            # Invert relationship and change type to be semantically correct
            # Input: dep_pkg DEPENDENCY_OF package (dep_pkg is a dependency OF package)
            # Output: package DEPENDS_ON dep_pkg (package depends on dep_pkg)
            key = (rel.to_, "depends-on")
            grouped.setdefault(key, set()).add(rel.from_)
        elif rel.type.value == "described-by":
            # from describes to (lock → non-lock)
            key = (rel.from_, rel.type.value)
            grouped.setdefault(key, set()).add(rel.to_)
    result: list[dict[str, str | list[str]]] = [
        {"from": from_ref, "to": list(deps), "type": rel_type}
        for (from_ref, rel_type), deps in grouped.items()
    ]

    return result


def build_sbom_metadata(namespace: str, version: str | None, timestamp: str) -> dict[str, Any]:
    return {
        "name": namespace,
        "version": version,
        "timestamp": timestamp,
        "tool": "Fluid-Labels",
        "organization": "Fluid attacks",
    }



import base64
import json
import logging
import re
import shutil
import subprocess
import tempfile
from collections.abc import Iterator
from contextlib import contextmanager
from pathlib import Path
from typing import Any

from labels.model.sources import ImageContext, ImageMetadata, LayerData
from labels.model.syft_sbom import (
    SyftAnnotation,
    SyftArtifact,
    SyftLocation,
    SyftMetadata,
    SyftRelationship,
    SyftSBOM,
    SyftSource,
)
from labels.utils.exceptions import (
    DockerImageNotFoundError,
    InvalidImageReferenceError,
    SkopeoNotFoundError,
    SyftNotFoundError,
)
from labels.utils.file import extract_tar_file

LOGGER = logging.getLogger(__name__)

ADDITIONAL_CATALOGERS = [
    "dart-pubspec-lock-cataloger",
    "dart-pubspec-cataloger",
    "swift-package-manager-cataloger",
    "cocoapods-cataloger",
]


def _format_image_ref(image_ref: str, *, daemon: bool = False) -> str:
    image_ref_pattern = (
        r"^(?:(?P<host>[\w\.\-]+(?:\:\d+)?)/)?"
        r"(?P<namespace>(?:[\w\.\-]+(?:/[\w\.\-]+)*)?/)?"
        r"(?P<image>[\w\.\-]+)(?::(?P<tag>[\w\.\-]+))?(?:@"
        r"(?P<digest>sha256:[A-Fa-f0-9]{64}))?$"
    )
    prefix_to_use = "docker-daemon:" if daemon else "docker://"
    prefix_used: str | None = None
    prefixes = ["docker://", "docker-daemon:"]
    for prefix in prefixes:
        if image_ref.startswith(prefix):
            image_ref = image_ref.replace(prefix, "", 1)
            prefix_used = prefix
            break

    prefix_to_use = prefix_used or prefix_to_use

    if re.match(image_ref_pattern, image_ref):
        return f"{prefix_to_use}{image_ref}"

    raise InvalidImageReferenceError(image_ref)


def _get_skopeo_path() -> str:
    skopeo_path = shutil.which("skopeo")
    if not skopeo_path:
        raise SkopeoNotFoundError
    return skopeo_path


def _execute_command(command_args: list[str]) -> bool:
    with subprocess.Popen(  # noqa: S603
        command_args,
        shell=False,
        stdout=subprocess.PIPE,
    ) as proc:
        exit_code = proc.wait()
        return exit_code == 0


def _load_manifest(layers_dir: str) -> dict[str, Any]:
    path = Path(layers_dir, "manifest.json")
    with path.open(encoding="utf-8") as f:
        return json.load(f)


def _load_config(config_digest: str, layers_dir: str) -> dict[str, Any]:
    path = Path(layers_dir, config_digest.replace("sha256:", ""))
    with path.open(encoding="utf-8") as f:
        return json.load(f)


def _extract_layer(layer: dict[str, Any], layers_dir: str, output_dir: str) -> None:
    layer_digest = layer["digest"].replace("sha256:", "")
    tar_path = Path(layers_dir, layer_digest)
    if tar_path.exists():
        dest = Path(output_dir, layer["digest"])
        dest.mkdir(parents=True, exist_ok=True)
        extract_tar_file(str(tar_path), str(dest))


def _build_copy_auth_args(
    *,
    unauthenticated_args: list[str],
    username: str | None = None,
    password: str | None = None,
    token: str | None = None,
    aws_creds: str | None = None,
) -> list[str]:
    if username and password:
        unauthenticated_args.extend(
            ["--src-username", username, "--src-password", password],
        )

    elif token:
        unauthenticated_args.extend(["--src-registry-token", token])

    elif aws_creds:
        unauthenticated_args.append(f"--src-creds={aws_creds}")

    return unauthenticated_args


def _build_inspect_auth_args(
    *,
    unauthenticated_args: list[str],
    username: str | None = None,
    password: str | None = None,
    token: str | None = None,
    aws_creds: str | None = None,
) -> list[str]:
    if username and password:
        unauthenticated_args.extend(["--username", username, "--password", password])

    elif token:
        unauthenticated_args.append(f"--registry-token={token}")

    elif aws_creds:
        unauthenticated_args.append(f"--creds={aws_creds}")

    return unauthenticated_args


def _custom_object_hook(json_object: dict[str, Any]) -> ImageMetadata | dict:
    if "Name" in json_object and "Digest" in json_object and "RepoTags" in json_object:
        layersdata = [
            LayerData(
                mimetype=layer_data["MIMEType"],
                digest=layer_data["Digest"],
                size=layer_data["Size"],
                annotations=layer_data.get("Annotations"),
            )
            for layer_data in json_object["LayersData"]
        ]
        return ImageMetadata(
            name=json_object["Name"],
            digest=json_object["Digest"],
            repotags=json_object["RepoTags"],
            created=json_object["Created"],
            dockerversion=json_object["DockerVersion"],
            labels=json_object["Labels"],
            architecture=json_object["Architecture"],
            os=json_object["Os"],
            layers=json_object["Layers"],
            layersdata=layersdata,
            env=json_object["Env"],
        )
    return json_object


def copy_image(  # noqa: PLR0913
    image_ref: str,
    dest_path: str,
    *,
    username: str | None = None,
    password: str | None = None,
    aws_creds: str | None = None,
    token: str | None = None,
    os_override: str | None = None,
    arch_override: str | None = None,
) -> bool:
    skopeo_path = _get_skopeo_path()

    formated_image_ref = _format_image_ref(image_ref)

    command_args = [
        skopeo_path,
        "copy",
        "--dest-decompress",
        "--insecure-policy",
        *build_override_args(os_override=os_override, arch_override=arch_override),
        formated_image_ref,
        f"dir:{dest_path}",
    ]

    authenticated_args = _build_copy_auth_args(
        unauthenticated_args=command_args,
        username=username,
        password=password,
        aws_creds=aws_creds,
        token=token,
    )

    return _execute_command(authenticated_args)


def extract_docker_image(  # noqa: PLR0913
    image: ImageMetadata,
    output_dir: str,
    *,
    username: str | None = None,
    password: str | None = None,
    os_override: str | None = None,
    arch_override: str | None = None,
    token: str | None = None,
    aws_creds: str | None = None,
    daemon: bool = False,
) -> tuple[str, dict[str, Any]]:
    layers_dir_temp = tempfile.mkdtemp()

    formated_image_ref = _format_image_ref(image.image_ref, daemon=daemon)

    copy_image(
        image_ref=formated_image_ref or image.image_ref,
        dest_path=layers_dir_temp,
        username=username,
        password=password,
        os_override=os_override,
        arch_override=arch_override,
        token=token,
        aws_creds=aws_creds,
    )

    manifest = _load_manifest(layers_dir_temp)
    manifest["config_full"] = _load_config(manifest["config"]["digest"], layers_dir_temp)

    for layer in manifest["layers"]:
        _extract_layer(layer, layers_dir_temp, output_dir)

    return layers_dir_temp, manifest


def build_override_args(
    *,
    os_override: str | None = None,
    arch_override: str | None = None,
) -> list[str]:
    os_value = os_override if os_override is not None else "linux"
    override_args = []
    if os_value:
        override_args.append(f"--override-os={os_value}")
    if arch_override:
        override_args.append(f"--override-arch={arch_override}")
    return override_args


def get_docker_image(  # noqa: PLR0913
    image_ref: str,
    *,
    username: str | None = None,
    password: str | None = None,
    os_override: str | None = None,
    arch_override: str | None = None,
    token: str | None = None,
    aws_creds: str | None = None,
    daemon: bool = False,
) -> ImageMetadata:
    skopeo_path = _get_skopeo_path()

    formated_image_ref = _format_image_ref(image_ref, daemon=daemon)

    command_args = [
        skopeo_path,
        "inspect",
        *build_override_args(os_override=os_override, arch_override=arch_override),
        formated_image_ref,
    ]
    authenticated_args = _build_inspect_auth_args(
        unauthenticated_args=command_args,
        username=username,
        password=password,
        aws_creds=aws_creds,
        token=token,
    )

    try:
        result = subprocess.run(  # noqa: S603
            authenticated_args,
            check=True,
            capture_output=True,
            text=True,
        )
        image_metadata: ImageMetadata = json.loads(
            result.stdout,
            object_hook=_custom_object_hook,
        )
        image_metadata = (
            image_metadata.model_copy(
                update={"image_ref": image_ref},
            )
            if image_metadata
            else image_metadata
        )
    except subprocess.CalledProcessError as error:
        raise DockerImageNotFoundError(image_ref, error.stderr.strip()) from error
    else:
        return image_metadata


def get_image_context(  # noqa: PLR0913
    *,
    image: ImageMetadata,
    username: str | None = None,
    password: str | None = None,
    os_override: str | None = None,
    arch_override: str | None = None,
    token: str | None = None,
    aws_creds: str | None = None,
    daemon: bool = False,
) -> ImageContext:
    temp_dir = tempfile.mkdtemp()

    layers_dir, manifest = extract_docker_image(
        image,
        temp_dir,
        username=username,
        password=password,
        os_override=os_override,
        arch_override=arch_override,
        token=token,
        aws_creds=aws_creds,
        daemon=daemon,
    )

    return ImageContext(
        id=image.digest,
        name=image.name,
        publisher="",
        arch=image.architecture,
        size=str(sum(x.size for x in image.layersdata)),
        full_extraction_dir=temp_dir,
        layers_dir=layers_dir,
        manifest=manifest,
        image_ref=image.image_ref,
    )


def _get_syft_path() -> str:
    syft_path = shutil.which("syft")
    if not syft_path:
        raise SyftNotFoundError
    return syft_path


def _build_platform_arg(
    *,
    os_override: str | None = None,
    arch_override: str | None = None,
) -> list[str]:
    os_value = os_override if os_override else "linux"
    arch_override_value = arch_override if arch_override else "arm64"
    return ["--platform", f"{os_value}/{arch_override_value}"]


def _get_formatted_image(*, image_ref: str, daemon: bool = False) -> str:
    image_ref_pattern = (
        r"^(?:(?P<host>[\w\.\-]+(?:\:\d+)?)/)?"
        r"(?P<namespace>(?:[\w\.\-]+(?:/[\w\.\-]+)*)?/)?"
        r"(?P<image>[\w\.\-]+)(?::(?P<tag>[\w\.\-]+))?(?:@"
        r"(?P<digest>sha256:[A-Fa-f0-9]{64}))?$"
    )

    syft_prefix = "docker" if daemon else "registry"
    prefixes = {"registry": "docker://", "docker": "docker-daemon:"}
    for prefix, prefix_value in prefixes.items():
        if image_ref.startswith(prefix_value):
            image_ref = image_ref.replace(prefix_value, "", 1)
            syft_prefix = prefix
            break

    if re.match(image_ref_pattern, image_ref):
        return f"{syft_prefix}:{image_ref}"

    raise InvalidImageReferenceError(image_ref)


@contextmanager
def _syft_auth_env(
    *,
    image_ref: str | None = None,
    username: str | None = None,
    password: str | None = None,
    aws_creds: str | None = None,
    token: str | None = None,
) -> Iterator[dict[str, str]]:
    if username and password:
        LOGGER.info("Setting DOCKER_USERNAME and DOCKER_PASSWORD environment variables for Syft")
        yield {
            "SYFT_REGISTRY_AUTH_USERNAME": username,
            "SYFT_REGISTRY_AUTH_PASSWORD": password,
        }
        return

    if image_ref and (auth := token or aws_creds):
        config_dir = tempfile.mkdtemp(prefix=".syft")
        LOGGER.debug("Creating temporary Docker config dir at %s", config_dir)

        try:
            docker_config_path = Path(config_dir, "config.json")
            registry_url = image_ref.split("/", 1)[0]

            encoded_auth = base64.b64encode(auth.encode("utf-8")).decode("utf-8")
            docker_config = {
                "auths": {
                    registry_url: {
                        "auth": encoded_auth,
                    },
                }
            }

            with docker_config_path.open("w", encoding="utf-8") as f:
                json.dump(docker_config, f)
            yield {
                "DOCKER_CONFIG": config_dir,
            }
        finally:
            LOGGER.debug("Cleaning up temporary DOCKER_CONFIG directory: %s", config_dir)
            shutil.rmtree(config_dir, ignore_errors=True)

        return

    yield {}


def _get_metadata(metadata: dict[str, Any]) -> SyftMetadata:
    dependencies = (
        metadata.get("depends", [])
        if metadata.get("depends")
        else metadata.get("pullDependencies", [])
    )
    release = metadata.get("release", "") if isinstance(metadata.get("release"), str) else ""
    package = str(metadata.get("package")) if metadata.get("package") else "UNKNOWN"
    version = str(metadata.get("version")) if metadata.get("version") else "UNKNOWN"
    source = metadata.get("source", "") if isinstance(metadata.get("source"), str) else ""

    return SyftMetadata(
        package=package,
        version=version,
        provides=metadata.get("provides", []),
        dependencies=dependencies,
        maintainer=metadata.get("maintainer", ""),
        origin_package=metadata.get("originPackage", ""),
        architecture=metadata.get("architecture", ""),
        source=source,
        source_version=metadata.get("sourceVersion", ""),
        pre_dependencies=metadata.get("preDepends", []),
        name=metadata.get("name", ""),
        epoch=metadata.get("epoch"),
        release=release,
        source_rpm=metadata.get("sourceRpm", ""),
        base_package=metadata.get("basepackage", ""),
        packager=metadata.get("packager", ""),
        requires_dist=metadata.get("requiresDist", []),
        provides_extra=metadata.get("providesExtra", []),
        manifest=metadata.get("manifest", {}),
        pom_properties=metadata.get("pomProperties", {}),
        pom_project=metadata.get("pomProject", {}),
    )


def _get_syft_source(source: dict[str, Any]) -> SyftSource:
    namespace = str(source.get("metadata", {}).get("userInput"))
    version = ""

    if (digests := source.get("metadata", {}).get("repoDigests")) and (first_digest := digests[0]):
        version = str(first_digest).split("@", 1)[1]

    return SyftSource(
        namespace=namespace,
        version=version,
    )


def _get_syft_relationships(relationships: list[dict[str, Any]]) -> list[SyftRelationship]:
    return [
        SyftRelationship(
            parent=parent,
            child=child,
            type=type_,
        )
        for relationship in relationships
        if (parent := relationship.get("parent"))
        and (child := relationship.get("child"))
        and (type_ := relationship.get("type"))
    ]


def _get_syft_artifacts(artifacts: list[dict[str, Any]]) -> list[SyftArtifact]:
    return [
        SyftArtifact(
            id_=str(artifact.get("id")),
            name=str(artifact.get("name")),
            version=str(artifact.get("version")),
            type=str(artifact.get("type")),
            locations=[
                SyftLocation(
                    path=location.get("path"),
                    access_path=location.get("path"),
                    layer_id=location.get("layerID"),
                    annotations=(
                        SyftAnnotation(
                            evidence=location.get("annotations", {}).get("evidence", ""),
                        )
                        if location.get("annotations")
                        else None
                    ),
                )
                for location in artifact.get("locations", [])
            ],
            p_url=str(artifact.get("purl")),
            metadata=_get_metadata(artifact.get("metadata", {})),
            found_by=str(artifact.get("foundBy")),
            language=str(artifact.get("language")),
        )
        for artifact in artifacts
        if artifact.get("id")
        and artifact.get("name")
        and artifact.get("version")
        and artifact.get("version") != "UNKNOWN"
    ]


def _get_syft_object_hook(json_object: dict[str, Any]) -> SyftSBOM | dict:
    if sbom_artifacts := json_object.get("artifacts"):
        sbom_source = json_object.get("source", {})
        sbom_relationships = json_object.get("artifactRelationships", [])
        return SyftSBOM(
            artifacts=_get_syft_artifacts(sbom_artifacts),
            source=_get_syft_source(sbom_source),
            relationships=_get_syft_relationships(sbom_relationships),
        )

    return json_object


def get_docker_sbom(  # noqa: PLR0913
    image_ref: str,
    *,
    username: str | None = None,
    password: str | None = None,
    os_override: str | None = None,
    arch_override: str | None = None,
    token: str | None = None,
    aws_creds: str | None = None,
    daemon: bool = False,
) -> SyftSBOM:
    syft_path = _get_syft_path()
    formatted_img = _get_formatted_image(image_ref=image_ref, daemon=daemon)

    LOGGER.info("Retrieving SBOM for image: %s", image_ref)
    LOGGER.info("Retrieving SBOM for image: %s", formatted_img)

    command_args = [
        syft_path,
        "scan",
        formatted_img,
        *_build_platform_arg(os_override=os_override, arch_override=arch_override),
        "--select-catalogers",
        ",".join("+" + cataloger for cataloger in ADDITIONAL_CATALOGERS),
        "-o",
        "json",
    ]

    with _syft_auth_env(
        image_ref=image_ref,
        username=username,
        password=password,
        aws_creds=aws_creds,
        token=token,
    ) as syft_envs:
        try:
            result = subprocess.run(  # noqa: S603
                command_args,
                check=True,
                capture_output=True,
                text=True,
                env=syft_envs,
            )

            sbom_data: SyftSBOM = json.loads(
                result.stdout,
                object_hook=_get_syft_object_hook,
            )
        except subprocess.CalledProcessError as error:
            raise DockerImageNotFoundError(image_ref, error.stderr.strip()) from error

    return sbom_data



from typing import cast

from labels.model.file import LocationReadCloser
from labels.model.indexables import IndexedDict, IndexedList, ParsedValue
from labels.model.package import Package
from labels.model.relationship import Relationship, RelationshipType
from labels.model.release import Environment
from labels.model.resolver import Resolver
from labels.parsers.cataloger.php.package_builder import new_package_from_composer
from labels.parsers.cataloger.utils import get_enriched_location
from labels.parsers.collection.json import parse_json_with_tree_sitter

EMPTY_DICT: IndexedDict[str, ParsedValue] = IndexedDict()


def parse_composer_lock(
    _resolver: Resolver | None,
    _environment: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    file_content = parse_json_with_tree_sitter(reader.read_closer.read())
    if not isinstance(file_content, IndexedDict):
        return [], []

    packages = _collect_packages(file_content, reader)
    relationships = _collect_relationships(file_content, packages)

    return packages, relationships


def _collect_packages(
    file_content: IndexedDict[str, ParsedValue], reader: LocationReadCloser
) -> list[Package]:
    packages: list[Package] = []

    for is_dev, raw_package in _get_all_packages_with_dev_flag(file_content):
        name = str(raw_package.get("name", "")) or None
        version = str(raw_package.get("version", "")) or None

        if not name:
            continue

        new_location = get_enriched_location(
            reader.location,
            line=raw_package.get_key_position("name").start.line,
            is_dev=is_dev,
            is_transitive=False,
        )

        package = new_package_from_composer(name=name, version=version, location=new_location)
        if package:
            packages.append(package)

    return packages


def _collect_relationships(
    package_json: IndexedDict[str, ParsedValue], packages: list[Package]
) -> list[Relationship]:
    relationships: list[Relationship] = []

    all_packages = [package for _, package in _get_all_packages_with_dev_flag(package_json)]

    for raw_package in all_packages:
        name = str(raw_package.get("name", "")) or None
        if not name:
            continue

        parsed_package = next((x for x in packages if x.name == name), None)
        if not parsed_package:
            continue

        if isinstance(raw_package.get("require"), IndexedDict):
            require = cast("dict[str, str]", raw_package.get("require"))
            deps = list(require.keys())

            for dep_name in deps:
                package_dep = next((x for x in packages if x.name == dep_name), None)
                if package_dep:
                    relationships.append(
                        Relationship(
                            from_=parsed_package.id_,
                            to_=package_dep.id_,
                            type=RelationshipType.DEPENDENCY_OF_RELATIONSHIP,
                        ),
                    )

    return relationships


def _get_all_packages_with_dev_flag(
    package_json: IndexedDict[str, ParsedValue],
) -> list[tuple[bool, IndexedDict[str, ParsedValue]]]:
    packages: list[tuple[bool, IndexedDict[str, ParsedValue]]] = []

    production_packages = package_json.get("packages")
    development_packages = package_json.get("packages-dev")

    if isinstance(production_packages, IndexedList):
        packages.extend(
            [
                (False, package)
                for package in production_packages
                if isinstance(package, IndexedDict)
            ]
        )

    if isinstance(development_packages, IndexedList):
        packages.extend(
            [
                (True, package)
                for package in development_packages
                if isinstance(package, IndexedDict)
            ]
        )

    return packages
